from logging import debug
import pytorch_lightning as pl
import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
import os
import json
from collections import defaultdict as ddict
from IPython import embed
from .BaseLitModel import BaseLitModel
from neuralkg.eval_task import *
from IPython import embed

from functools import partial

class KBATLitModel(BaseLitModel):
    def __init__(self, model, args):
        super().__init__(model, args)
        
    def forward(self, x):
        return self.model(x)
    
    def training_step(self, batch, batch_idx):
        num_epoch  = self.current_epoch
        if num_epoch < 3000: 
            model   = "GAT"
            adj     = batch['adj_matrix']
            n_hop   = batch['n_hop']
            pos_triple = batch['triples_GAT_pos']
            neg_triple = batch['triples_GAT_neg']
            pos_score  = self.model(pos_triple, model, adj, n_hop)
            neg_score  = self.model(neg_triple, model, adj, n_hop)
            loss       = self.loss(model, pos_score, neg_score)
        else:
            model   = "ConvKB"
            triples = batch['triples_Con']
            label   = batch['label']
            score   = self.model(triples, model)
            loss    = self.loss(model, score, label=label)
        self.log("Train|loss", loss,  on_step=False, on_epoch=True)
        return loss

    def validation_step(self, batch, batch_idx):
        # pos_triple, tail_label, head_label = batch
        results = dict()
        ranks = link_predict(batch, self.model)
        results["count"] = torch.numel(ranks)
        results["Eval|mrr"] = torch.sum(1.0 / ranks).item()
        for k in [1, 3, 10]:
            results['Eval|hits@{}'.format(k)] = torch.numel(ranks[ranks <= k])
        return results
    
    def validation_epoch_end(self, results) -> None:
        outputs = self.get_results(results, "Eval|")
        # self.log("Eval|mrr", outputs["Eval|mrr"], on_epoch=True)
        self.log_dict(outputs, prog_bar=True, on_epoch=True)

    def test_step(self, batch, batch_idx):
        results = dict()
        ranks = link_predict(batch, self.model)
        results["count"] = torch.numel(ranks)
        results["Test|mrr"] = torch.sum(1.0 / ranks).item()
        for k in [1, 3, 10]:
            results['Test|hits@{}'.format(k)] = torch.numel(ranks[ranks <= k])
        return results
    
    def test_epoch_end(self, results) -> None:
        outputs = self.get_results(results, "Test|")
        self.log_dict(outputs, prog_bar=True, on_epoch=True)
    
    def get_results(self, results, mode):
        outputs = ddict(float)
        count = np.array([o["count"] for o in results]).sum().item()
        metrics = ["mrr", "hits@1", "hits@3", "hits@10"]
        metrics = [mode + metric for metric in metrics]
        for metric in metrics:
            number = np.array([o[metric] for \
             o in results]).sum().item() / count
            outputs[metric] = round(number, 2)
        return outputs

    '''这里设置优化器和lr_scheduler'''
    def configure_optimizers(self):
        if self.current_epoch < 3000:
            optimizer = self.optimizer_class(self.model.parameters(), lr=self.args.lr, weight_decay=1e-6)
            StepLR = torch.optim.lr_scheduler.StepLR(optimizer, step_size=500, gamma=0.5, last_epoch=-1)

        else:
            optimizer = self.optimizer_class(self.model.parameters(), lr=self.args.lr, weight_decay=1e-5)
            StepLR = torch.optim.lr_scheduler.StepLR(optimizer, step_size=25, gamma=0.5, last_epoch=-1)

        optim_dict = {'optimizer': optimizer, 'lr_scheduler': StepLR}
        return optim_dict



import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from torch.autograd import Variable
import time
import os

class KBAT(nn.Module):

    """`Learning Attention-based Embeddings for Relation Prediction in Knowledge Graphs`_ (KBAT), 
        which introduces the attention to aggregate the neighbor node representation.

    Attributes:
        args: Model configuration parameters.
    
    .. _Learning Attention-based Embeddings for Relation Prediction in Knowledge Graphs: 
        https://arxiv.org/pdf/1906.01195.pdf
    """

    def __init__(self, args):
        super(KBAT,self).__init__()
        self.args = args
        self.entity_embeddings   = None
        self.relation_embeddings = None

        self.init_GAT_emb()
        self.init_ConvKB_emb()

    def init_GAT_emb(self):
        """Initialize the GAT model and embeddings 

        Args:
            ent_emb_out: Entity embedding, shape:[num_ent, emb_dim].
            rel_emb_out: Relation_embedding, shape:[num_rel, emb_dim].
            entity_embeddings: The final embedding used in ConvKB.
            relation_embeddings: The final embedding used in ConvKB.
            attentions, out_att: The graph attention layers.
        """
        self.num_ent = self.args.num_ent
        self.num_rel = self.args.num_rel
        self.emb_dim = self.args.emb_dim

        self.ent_emb_out = nn.Parameter(torch.randn(self.num_ent,self.emb_dim))
        self.rel_emb_out = nn.Parameter(torch.randn(self.num_rel,self.emb_dim))

        self.drop  = 0.3
        self.alpha = 0.2  

        self.nheads_GAT = 2
        self.out_dim = 100

        self.entity_embeddings = nn.Parameter(
            torch.randn(self.num_ent, self.out_dim * self.nheads_GAT))

        self.relation_embeddings = nn.Parameter(
            torch.randn(self.num_rel, self.out_dim * self.nheads_GAT))

        self.dropout_layer = nn.Dropout(self.drop)

        self.attentions = [GraphAttentionLayer(self.num_ent, 
                                                 self.emb_dim,
                                                 self.out_dim,
                                                 self.emb_dim,
                                                 dropout=self.drop,
                                                 alpha=self.alpha,
                                                 concat=True)
                           for _ in range(self.nheads_GAT)] 

        for i, attention in enumerate(self.attentions):
            self.add_module('attention_{}'.format(i), attention)

        # W matrix to convert h_input to h_output dimension 变换矩阵
        self.W = nn.Parameter(torch.zeros(
            size=(self.emb_dim, self.nheads_GAT * self.out_dim)))
        nn.init.xavier_uniform_(self.W.data, gain=1.414)

        self.out_att = GraphAttentionLayer(self.num_ent, 
                                             self.out_dim * self.nheads_GAT,
                                             self.out_dim * self.nheads_GAT,
                                             self.out_dim * self.nheads_GAT,
                                             dropout=self.drop,
                                             alpha=self.alpha,
                                             concat=False
                                             )

        self.W_entities = nn.Parameter(torch.zeros(
            size=(self.emb_dim, self.out_dim * self.nheads_GAT)))
        nn.init.xavier_uniform_(self.W_entities.data, gain=1.414)

    def init_ConvKB_emb(self): 
        """Initialize the ConvKB model.

        Args:
            conv_layer: The convolution layer.
            dropout: The dropout layer.
            ReLU: Relu activation function.
            fc_layer: The full connection layer.
        """
        self.conv_layer = nn.Conv2d(1, 50, (1,3))  
        self.dropout    = nn.Dropout(0.3)
        self.ReLU       = nn.ReLU()
        self.fc_layer   = nn.Linear(10000, 1)

        nn.init.xavier_uniform_(self.fc_layer.weight, gain=1.414)
        nn.init.xavier_uniform_(self.conv_layer.weight, gain=1.414)

    def forward(self, triples, mode, adj_matrix=None, n_hop=None):
        """The functions used in the training and testing phase

        Args:
            triples: The triples ids, as (h, r, t), shape:[batch_size, 3].
            mode: The mode indicates that the model will be used, when it 
            is 'GAT', it means graph attetion model, when it is 'ConvKB', 
            it means ConvKB model. 

        Returns:
            score: The score of triples.
        """
        if mode == 'GAT':  # gat
            score = self.forward_GAT(triples, adj_matrix, n_hop)

        else:
            score = self.forward_Con(triples, mode)
        return score

    def forward_Con(self, triples, mode):
        score = None

        if mode == 'ConvKB':
            head_emb = self.entity_embeddings[triples[:, 0]].unsqueeze(1)
            rela_emb = self.relation_embeddings[triples[:, 1]].unsqueeze(1) 
            tail_emb = self.entity_embeddings[triples[:, 2]].unsqueeze(1)
            score = self.cal_Con_score(head_emb, rela_emb, tail_emb)
                
        elif mode == 'head-batch':
            head_emb = self.entity_embeddings.unsqueeze(1)  # [1, num_ent, dim]
            for triple in triples:
                rela_emb = self.relation_embeddings[triple[1]].\
                                unsqueeze(0).tile(dims=(self.num_ent,1,1))
                tail_emb = self.entity_embeddings[triple[2]].\
                                unsqueeze(0).tile(dims=(self.num_ent,1,1))
                s = self.cal_Con_score(head_emb, rela_emb, tail_emb).t()
                
                if score == None:
                    score = s
                else:
                    score = torch.cat((score, s), dim=0)


        elif mode == 'tail-batch':
            tail_emb = self.entity_embeddings.unsqueeze(1)  # [1, num_ent, dim]
            for triple in triples:
                head_emb = self.entity_embeddings[triple[0]].\
                                unsqueeze(0).tile(dims=(self.num_ent,1,1))
                rela_emb = self.relation_embeddings[triple[1]].\
                                unsqueeze(0).tile(dims=(self.num_ent,1,1))
                s = self.cal_Con_score(head_emb, rela_emb, tail_emb).t()
                
                if score == None:
                    score = s
                else:
                    score = torch.cat((score, s), dim=0)

        return score

    def forward_GAT(self, triples, adj_matrix, n_hop):
        edge_list = adj_matrix[0] #边节点
        edge_type = adj_matrix[1] #边种类
        edge_list_nhop = torch.cat((n_hop[:, 3].unsqueeze(-1), 
                                        n_hop[:, 0].unsqueeze(-1)), dim=1).t()
        edge_type_nhop = torch.cat([n_hop[:, 1].unsqueeze(-1), 
                                        n_hop[:, 2].unsqueeze(-1)], dim=1)

        edge_emb = self.rel_emb_out[edge_type]
        self.ent_emb_out.data = F.normalize(self.ent_emb_out.data, p=2, dim=1).detach()
        edge_embed_nhop = self.rel_emb_out[edge_type_nhop[:, 0]] + \
                self.rel_emb_out[edge_type_nhop[:, 1]]
            
        ent_emb_out = torch.cat([att(self.ent_emb_out, edge_list, edge_emb, edge_list_nhop, 
                                edge_embed_nhop) for att in self.attentions], dim=1)
        ent_emb_out = self.dropout_layer(ent_emb_out)
        rel_emb_out = self.rel_emb_out.mm(self.W)
        edge_emb = rel_emb_out[edge_type]
        edge_embed_nhop = rel_emb_out[edge_type_nhop[:, 0]] + \
                rel_emb_out[edge_type_nhop[:, 1]]
        ent_emb_out = F.elu(self.out_att(ent_emb_out, edge_list, edge_emb,
                                edge_list_nhop, edge_embed_nhop))
            
        mask_indices = torch.unique(triples[:, 2])
        mask = torch.zeros(self.ent_emb_out.shape[0]).type_as(self.ent_emb_out)
        mask[mask_indices] = 1.0

        entities_upgraded = self.ent_emb_out.mm(self.W_entities)
        ent_emb_out = entities_upgraded + \
                mask.unsqueeze(-1).expand_as(ent_emb_out) * ent_emb_out
        ent_emb_out = F.normalize(ent_emb_out, p=2, dim=1)

        self.entity_embeddings.data = ent_emb_out.data
        self.relation_embeddings.data = rel_emb_out.data

        head_emb = ent_emb_out[triples[:, 0]]
        rela_emb = rel_emb_out[triples[:, 1]]
        tail_emb = ent_emb_out[triples[:, 2]]
        return self.cal_GAT_score(head_emb, rela_emb, tail_emb)

    def cal_Con_score(self, head_emb, rela_emb, tail_emb):

        """Calculating the score of triples with ConvKB model.

        Args:
            head_emb: The head entity embedding.
            rela_emb: The relation embedding.
            tail_emb: The tail entity embedding.

        Returns:
            score: The score of triples.
        """      

        conv_input = torch.cat((head_emb, rela_emb, tail_emb), dim=1)            
        batch_size= conv_input.shape[0]

        conv_input = conv_input.transpose(1, 2)
        conv_input = conv_input.unsqueeze(1)
        
        out_conv = self.conv_layer(conv_input)
        out_conv = self.ReLU(out_conv)
        out_conv = self.dropout(out_conv)
        out_conv = out_conv.squeeze(-1).view(batch_size, -1)
        score = self.fc_layer(out_conv)
        return score

    def cal_GAT_score(self, head_emb, relation_emb, tail_emb):

        """Calculating the score of triples with TransE model.

        Args:
            head_emb: The head entity embedding.
            rela_emb: The relation embedding.
            tail_emb: The tail entity embedding.

        Returns:
            score: The score of triples.
        """

        score = (head_emb + relation_emb) - tail_emb
        score = torch.norm(score, p=1, dim=1)
        return score

class SpecialSpmmFunctionFinal(torch.autograd.Function):
    """
    Special function for only sparse region backpropataion layer, similar to https://arxiv.org/abs/1710.10903
    """
    @staticmethod
    def forward(ctx, edge, edge_w, N, E, out_features):

        a = torch.sparse_coo_tensor(
            edge, edge_w, torch.Size([N, N, out_features]))
        b = torch.sparse.sum(a, dim=1)
        ctx.N = b.shape[0]
        ctx.outfeat = b.shape[1]
        ctx.E = E
        ctx.indices = a._indices()[0, :]

        return b.to_dense()

    @staticmethod
    def backward(ctx, grad_output):
        grad_values = None
        if ctx.needs_input_grad[1]:
            edge_sources = ctx.indices

            grad_values = grad_output[edge_sources]

        return None, grad_values, None, None, None

class SpecialSpmmFinal(nn.Module):
    """
    Special spmm final layer, similar to https://arxiv.org/abs/1710.10903.
    """
    def forward(self, edge, edge_w, N, E, out_features):
        return SpecialSpmmFunctionFinal.apply(edge, edge_w, N, E, out_features)

class GraphAttentionLayer(nn.Module):
    
    """
    Sparse version GAT layer, similar to https://arxiv.org/abs/1710.10903.
    """

    def __init__(self, num_nodes, in_features, out_features, nrela_dim, dropout, alpha, concat=True):
        super(GraphAttentionLayer, self).__init__()
        self.in_features = in_features   
        self.out_features = out_features 
        self.num_nodes = num_nodes       
        self.alpha = alpha        
        self.concat = concat
        self.nrela_dim = nrela_dim       

        self.a = nn.Parameter(torch.zeros(
            size=(out_features, 2 * in_features + nrela_dim)))
        nn.init.xavier_normal_(self.a.data, gain=1.414)
        self.a_2 = nn.Parameter(torch.zeros(size=(1, out_features)))
        nn.init.xavier_normal_(self.a_2.data, gain=1.414)

        self.dropout = nn.Dropout(dropout)
        self.leakyrelu = nn.LeakyReLU(self.alpha)
        self.special_spmm_final = SpecialSpmmFinal()

    def forward(self, input, edge, edge_embed, edge_list_nhop, edge_embed_nhop):
        N = input.size()[0]

        # Self-attention on the nodes - Shared attention mechanism
        edge = torch.cat((edge[:, :], edge_list_nhop[:, :]), dim=1)
        edge_embed = torch.cat(
            (edge_embed[:, :], edge_embed_nhop[:, :]), dim=0)

        edge_h = torch.cat(
            (input[edge[0, :], :], input[edge[1, :], :], edge_embed[:, :]), dim=1).t()
        # edge_h: (2*in_dim + nrela_dim) x E

        edge_m = self.a.mm(edge_h)
        # edge_m: D * E

        # to be checked later
        powers = -self.leakyrelu(self.a_2.mm(edge_m).squeeze())
        edge_e = torch.exp(powers).unsqueeze(1)
        assert not torch.isnan(edge_e).any()
        # edge_e: E

        e_rowsum = self.special_spmm_final(
            edge, edge_e, N, edge_e.shape[0], 1)
        e_rowsum[e_rowsum == 0.0] = 1e-12

        e_rowsum = e_rowsum
        # e_rowsum: N x 1
        edge_e = edge_e.squeeze(1)

        edge_e = self.dropout(edge_e)
        # edge_e: E

        edge_w = (edge_e * edge_m).t()
        # edge_w: E * D

        h_prime = self.special_spmm_final(
            edge, edge_w, N, edge_w.shape[0], self.out_features)

        assert not torch.isnan(h_prime).any()
        # h_prime: N x out
        h_prime = h_prime.div(e_rowsum)
        # h_prime: N x out

        assert not torch.isnan(h_prime).any()
        if self.concat:
            # if this layer is not last layer,
            return F.elu(h_prime)
        else:
            # if this layer is last layer,
            return h_prime

    def __repr__(self):
        return self.__class__.__name__ + ' (' + str(self.in_features) + ' -> ' + str(self.out_features) + ')'


import torch
import os
from IPython import embed

def RGCN_predict(batch, model, predicion='all'):
    if predicion == 'all':
        tail_ranks = head_predict(batch, model)
        head_ranks = tail_predict(batch, model)
        ranks = torch.cat([tail_ranks,head_ranks])
    elif predicion == 'tail':
        ranks = head_predict(batch, model)
    elif predicion == 'head':
        ranks = tail_predict(batch, model)

    return ranks.float()

def head_predict(batch,model):
    pos_triple = batch['positive_sample']
    idx        = pos_triple[:, 0]
    label      = batch['head_label']
    graph      = batch['graph']
    ent        = batch['entity']
    rel        = batch['rela']
    norm       = batch['norm'] 
    return calc_ranks(graph, ent, rel, norm, pos_triple, idx, label, "head-batch", model)

def tail_predict(batch,model):
    pos_triple = batch['positive_sample']
    idx        = pos_triple[:, 2]
    label      = batch['tail_label']
    graph      = batch['graph']
    ent        = batch['entity']
    rel        = batch['rela']
    norm       = batch['norm'] 

    return calc_ranks(graph, ent, rel, norm, pos_triple, idx, label, "tail-batch", model)

def calc_ranks(graph, ent, rel, norm, pos_triple, idx, label, mode, model):
    pred = model(graph, ent, rel, norm, pos_triple, mode=mode)  #TODO： 这里怎么改合适？ 是重新写一个还是？
    b_range = torch.arange(pred.size()[0])
    target_pred = pred[b_range, idx]
    pred = torch.where(label.byte(), -torch.ones_like(pred) * 10000000, pred)
    pred[b_range, idx] = target_pred

    ranks = 1 + torch.argsort(torch.argsort(pred, dim=1, descending=True),
                                    dim=1, descending=False)[b_range, idx]
    return ranks


from .ComplEx import ComplEx
from .TransE import TransE
from .DistMult import DistMult
from .RotatE import RotatE
from .TransH import TransH
from .TransR import TransR
from .SimplE import SimplE
from .BoxE import BoxE
from .ConvE import ConvE
from .CrossE import CrossE



from logging import debug
import pytorch_lightning as pl
import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
import os
import json
from collections import defaultdict as ddict
from IPython import embed
from .BaseLitModel import BaseLitModel
from neuralkg.eval_task import *
from IPython import embed

import pickle
import time

from functools import partial

class IterELitModel(BaseLitModel):
    def __init__(self, model, args):
        super().__init__(model, args)
        self.epoch=0
    

    def forward(self, x):
        return self.model(x)
    
    def training_step(self, batch, batch_idx):
        pos_sample = batch["positive_sample"]
        neg_sample = batch["negative_sample"]
        mode = batch["mode"]
        pos_score = self.model(pos_sample)
        neg_score = self.model(pos_sample, neg_sample, mode)
        if self.args.use_weight:
            subsampling_weight = batch["subsampling_weight"]
            loss = self.loss(pos_score, neg_score, subsampling_weight)
        else:
            loss = self.loss(pos_score, neg_score)
        self.log("Train|loss", loss,  on_step=False, on_epoch=True)
        return loss
    
    def training_epoch_end(self, results):
        self.epoch+=1
        if self.epoch % self.args.update_axiom_per == 0 and self.epoch !=0:
        #if True:
                # axioms include probability for each axiom in axiom pool
                # order: ref, sym, tran, inver, sub, equi, inferC
                # update_axioms:
                #            1) calculate probability for each axiom in axiom pool with current embeddings
                #            2) update the valid_axioms
                axioms_probability = self.update_axiom()
                #self.model.update_train_triples(epoch = self.epoch, update_per= self.args.update_axiom_per)
                updated_train_data = self.model.update_train_triples(epoch = self.epoch, update_per= self.args.update_axiom_per)
                if updated_train_data:
                    self.trainer.datamodule.data_train=updated_train_data
                #print('axiom_probability: %s' % (axioms_probability))


    def update_axiom(self):
        time_s = time.time()
        axiom_pro = self.model.run_axiom_probability()
        time_e = time.time()
        print('calculate axiom score:', time_e -time_s)
        with open('./save_axiom_prob/axiom_prob.pickle', 'wb') as f: pickle.dump(axiom_pro, f, pickle.HIGHEST_PROTOCOL)
        with open('./save_axiom_prob/axiom_pools.pickle', 'wb') as f: pickle.dump(self.model.axiompool, f, pickle.HIGHEST_PROTOCOL)
        self.model.update_valid_axioms(axiom_pro)
        return self.model.run_axiom_probability()

    def validation_step(self, batch, batch_idx):
        # pos_triple, tail_label, head_label = batch
        results = dict()
        ranks = link_predict(batch, self.model)
        results["count"] = torch.numel(ranks)
        results["Eval|mrr"] = torch.sum(1.0 / ranks).item()
        for k in [1, 3, 10]:
            results['Eval|hits@{}'.format(k)] = torch.numel(ranks[ranks <= k])
        return results
    
    def validation_epoch_end(self, results) -> None:
        outputs = self.get_results(results, "Eval|")
        # self.log("Eval|mrr", outputs["Eval|mrr"], on_epoch=True)
        self.log_dict(outputs, prog_bar=True, on_epoch=True)

    def test_step(self, batch, batch_idx):
        results = dict()
        ranks = link_predict(batch, self.model)
        results["count"] = torch.numel(ranks)
        results["Test|mrr"] = torch.sum(1.0 / ranks).item()
        for k in [1, 3, 10]:
            results['Test|hits@{}'.format(k)] = torch.numel(ranks[ranks <= k])
        return results
    
    def test_epoch_end(self, results) -> None:
        outputs = self.get_results(results, "Test|")
        self.log_dict(outputs, prog_bar=True, on_epoch=True)
    
    def get_results(self, results, mode):
        outputs = ddict(float)
        count = np.array([o["count"] for o in results]).sum().item()
        metrics = ["mrr", "hits@1", "hits@3", "hits@10"]
        metrics = [mode + metric for metric in metrics]
        for metric in metrics:
            number = np.array([o[metric] for \
             o in results]).sum().item() / count
            outputs[metric] = round(number, 2)
        return outputs

    '''这里设置优化器和lr_scheduler'''
    def configure_optimizers(self):
        milestones = int(self.args.max_epochs / 2)
        optimizer = self.optimizer_class(self.model.parameters(), lr=self.args.lr)
        StepLR = torch.optim.lr_scheduler.MultiStepLR(optimizer, milestones=[milestones], gamma=0.1)
        optim_dict = {'optimizer': optimizer, 'lr_scheduler': StepLR}
        return optim_dict



from .DataPreprocess import KGData
import pdb
class GroundAllRules:
    def __init__(self, args):
        self.MapRelation2ID = {}
        self.MapEntity2ID = {}
        self.Relation2Tuple = {}
        self.MapID2Entity = {}
        self.MapID2Relation = {}
        self.TrainTriples = {}

        self.RelSub2Obj = {}
        self.MapVariable = {}

        self.args = args
        self.fnEntityIDMap = args.data_path + "/entities.dict"
        self.fnRelationIDMap = args.data_path + "/relations.dict"
        path_len = len(args.data_path.split('/'))
        self.fnRuleType = args.data_path + "/" + args.data_path.split('/')[path_len - 1] + "_rule"
        self.fnTrainingTriples = args.data_path + "/train.txt"
        self.fnOutout = args.data_path + "/groudings.txt"

    def PropositionalizeRule(self):
        self.kgData = KGData(self.args)
        self.readData(self.fnEntityIDMap, self.fnRelationIDMap, self.fnTrainingTriples)
        self.groundRule(self.fnRuleType, self.fnOutout)

    def readData(self, fnEntityIDMap, fnRelationIDMap, fnTrainingTriples):
        tokens = []
        self.MapEntity2ID = self.kgData.ent2id
        self.MapRelation2ID = self.kgData.rel2id
        self.TrainTriples = self.kgData.TrainTriples
        self.Relation2Tuple = self.kgData.Relation2Tuple
        self.RelSub2Obj = self.kgData.RelSub2Obj
        # with open(fnEntityIDMap, 'r', encoding='utf-8') as f:
        #     for line in f.readlines():
        #         line = line.strip('\n')
        #         tokens = line.split("\t")
        #         iEntityID = int(tokens[0])
        #         strValue = tokens[1]
        #         self.MapEntity2ID[strValue] = iEntityID
        #         self.MapID2Entity[iEntityID] = strValue
        #
        # with open(fnRelationIDMap, "r", encoding='utf-8') as f:
        #     for line in f.readlines():
        #         line = line.strip('\n')
        #         tokens = line.split("\t")
        #         iRelationID = int(tokens[0])
        #         strValue = tokens[1]
        #         self.MapRelation2ID[strValue] = iRelationID
        #         self.MapID2Relation[iRelationID] = strValue

        print("Start to load soft rules......")

        # with open(fnTrainingTriples, "r", encoding='utf-8') as f:
        #     for line in f.readlines():
        #         line = line.strip('\n')
        #         tokens = line.split("\t")
        #         iRelationID = self.MapRelation2ID[tokens[1]]
        #         strValue = tokens[0] + "#" + tokens[2]
        #         line = line.replace(tokens[0], str(self.MapEntity2ID[tokens[0]]))
        #         line = line.replace(tokens[1], str(self.MapRelation2ID[tokens[1]]))
        #         line = line.replace(tokens[2], str(self.MapEntity2ID[tokens[2]]))
        #         self.TrainTriples[line] = True
        #         if not iRelationID in self.Relation2Tuple:
        #             tmpLst = []
        #             tmpLst.append(strValue)
        #             self.Relation2Tuple[iRelationID] = tmpLst
        #         else:
        #             self.Relation2Tuple[iRelationID].append(strValue)

        # with open(fnTrainingTriples, "r", encoding='utf-8') as f:
        #     for line in f.readlines():
        #         line = line.strip('\n')
        #         tokens = line.split("\t")
        #         iRelationID = self.MapRelation2ID[tokens[1]]
        #         iSubjectID = self.MapEntity2ID[tokens[0]]
        #         iObjectID = self.MapEntity2ID[tokens[2]]
        #         tmpMap = {}
        #         tmpMap_in = {}
        #         if not iRelationID in self.RelSub2Obj:
        #             if not iSubjectID in tmpMap:
        #                 tmpMap_in.clear()
        #                 tmpMap_in[iObjectID] = True
        #                 tmpMap[iSubjectID] = tmpMap_in
        #             else:
        #                 tmpMap[iSubjectID][iObjectID] = True
        #             self.RelSub2Obj[iRelationID] = tmpMap
        #         else:
        #             tmpMap = self.RelSub2Obj[iRelationID]
        #             if not iSubjectID in tmpMap:
        #                 tmpMap_in.clear()
        #                 tmpMap_in[iObjectID] = True
        #                 tmpMap[iSubjectID] = tmpMap_in
        #             else:
        #                 tmpMap[iSubjectID][iObjectID] = True
        #             self.RelSub2Obj[iRelationID] = tmpMap  # 是不是应该要加？
        print("success")

    def groundRule(self, fnRuleType, fnOutput):
        print("Start to propositionalize soft rules......")
        writer = open(fnOutput, "w")
        tmpLst = {}
        with open(fnRuleType, "r", encoding='utf-8') as f:
            for line in f.readlines():
                line = line.strip()
                if line.startswith("?"):
                    bodys = line.split("=>")[0].strip().split("  ")
                    heads = line.split("=>")[1].strip().split("  ")
                if len(bodys) == 3:
                    bEntity1 = bodys[0]
                    iFstRelation = self.MapRelation2ID[bodys[1]]
                    bEntity2 = bodys[2]
                    bEntity3 = heads[0]
                    iSndRelation = self.MapRelation2ID[heads[1]]
                    bEntity4 = heads[2].split("\t")[0]
                    hEntity1 = heads[2].split("\t")[1]
                    confi = float(hEntity1)
                    iSize = len(self.Relation2Tuple[iFstRelation])

                    for i in range(0, iSize):
                        strValue = self.Relation2Tuple.get(iFstRelation)[i]
                        iSubjectID = self.MapEntity2ID[strValue.split("#")[0]]
                        iObjectID = self.MapEntity2ID[strValue.split("#")[1]]
                        self.MapVariable[bEntity1] = iSubjectID
                        self.MapVariable[bEntity2] = iObjectID
                        strKey = "(" + str(iSubjectID) + "\t" + str(iFstRelation) + "\t" + str(
                            iObjectID) + ")\t" + "(" + str(self.MapVariable[bEntity3]) + "\t" + str(
                            iSndRelation) + "\t" + str(self.MapVariable[bEntity4]) + ")"
                        strCons = str(self.MapVariable[bEntity3]) + "\t" + str(iSndRelation) + "\t" + str(
                            self.MapVariable[bEntity4])
                        if (not strKey in tmpLst) and (not strCons in self.TrainTriples):
                            writer.write("2\t" + str(strKey) + "\t" + str(confi) + "\n")
                            tmpLst[strKey] = True
                        writer.flush()
                        self.MapVariable.clear()
                if len(bodys) == 6:
                    bEntity1 = bodys[0].strip()
                    iFstRelation = self.MapRelation2ID[bodys[1].strip()]
                    bEntity2 = bodys[2].strip()
                    bEntity3 = bodys[3].strip()
                    iSndRelation = self.MapRelation2ID[bodys[4].strip()]
                    bEntity4 = bodys[5].strip()
                    hEntity1 = heads[0].strip()
                    iTrdRelation = self.MapRelation2ID[heads[1].strip()]
                    hEntity2 = heads[2].split("\t")[0].strip()
                    confidence = heads[2].split("\t")[1].strip()
                    confi = float(confidence)
                    mapFstRel = self.RelSub2Obj[iFstRelation]
                    mapSndRel = self.RelSub2Obj[iSndRelation]
                    for lstEntity1 in mapFstRel:
                        self.MapVariable[bEntity1] = lstEntity1
                        lstEntity2 = list(mapFstRel[lstEntity1].keys())
                        iFstSize = len(lstEntity2)

                        for iFstIndex in range(0, iFstSize):
                            iEntity2ID = lstEntity2[iFstIndex]
                            self.MapVariable[bEntity1] = lstEntity1
                            self.MapVariable[bEntity2] = iEntity2ID
                            lstEntity3 = []
                            if (bEntity3 in self.MapVariable) and (self.MapVariable[bEntity3] in mapSndRel):
                                lstEntity3.append(self.MapVariable[bEntity3])
                            else:
                                if not bEntity3 in self.MapVariable:
                                    lstEntity3 = list(mapSndRel.keys())

                            iSndSize = len(lstEntity3)

                            for iSndIndex in range(0, iSndSize):
                                iEntity3ID = lstEntity3[iSndIndex]
                                self.MapVariable[bEntity1] = lstEntity1
                                self.MapVariable[bEntity2] = iEntity2ID
                                self.MapVariable[bEntity3] = iEntity3ID
                                lstEntity4 = []
                                if (bEntity4 in self.MapVariable) and (
                                        self.MapVariable[bEntity4] in mapSndRel[iEntity3ID]):
                                    lstEntity4.append(self.MapVariable[bEntity4])
                                else:
                                    if not bEntity4 in self.MapVariable:
                                        lstEntity4 = list(mapSndRel[iEntity3ID].keys())

                                iTrdSize = len(lstEntity4)

                                for iTrdIndex in range(0, iTrdSize):
                                    iEntity4ID = lstEntity4[iTrdIndex]
                                    self.MapVariable[bEntity4] = iEntity4ID
                                    infer = str(self.MapVariable[hEntity1]) + "\t" + str(iTrdRelation) + "\t" + str(
                                        self.MapVariable[hEntity2])
                                    strKey = "(" + str(lstEntity1) + "\t" + str(iFstRelation) + "\t" + str(
                                        iEntity2ID) + ")\t(" + str(iEntity3ID) + "\t" + str(iSndRelation) + "\t" + str(
                                        iEntity4ID) + ")\t" + "(" + str(self.MapVariable[hEntity1]) + "\t" + str(
                                        iTrdRelation) + "\t" + str(self.MapVariable[hEntity2]) + ")"
                                    if (not strKey in tmpLst) and (not infer in self.TrainTriples):
                                        writer.write("3\t" + strKey + "\t" + str(confi) + "\n")
                                        tmpLst[strKey] = True
                                self.MapVariable.clear()
                            self.MapVariable.clear()
                        writer.flush()
                        self.MapVariable.clear()



import torch.nn as nn
import torch
import torch.nn.functional as F
from .model import Model
from IPython import embed


class TransR(Model):
    """Learning Entity and Relation Embeddings for Knowledge Graph Completion`_ (TransR), which building entity and relation embeddings in separate entity space and relation spaces

    Attributes:
        args: Model configuration parameters.
        epsilon: Calculate embedding_range.
        margin: Calculate embedding_range and loss.
        embedding_range: Uniform distribution range.
        ent_emb: Entity embedding, shape:[num_ent, emb_dim].
        rel_emb: Relation embedding, shape:[num_rel, emb_dim].
        transfer_matrix: Transfer entity and relation embedding, shape:[num_rel, emb_dim*emb_dim]

    .. _Translating Embeddings for Modeling Multi-relational Data: http://www.aaai.org/ocs/index.php/AAAI/AAAI15/paper/download/9571/9523/
    """

    def __init__(self, args):
        super(TransR, self).__init__(args)
        self.args = args
        self.ent_emb = None
        self.rel_emb = None
        self.norm_flag = args.norm_flag
        self.init_emb()

    def init_emb(self):
        self.epsilon = 2.0
        self.margin = nn.Parameter(
            torch.Tensor([self.args.margin]), requires_grad=False
        )
        self.embedding_range = nn.Parameter(
            torch.Tensor([(self.margin.item() + self.epsilon) / self.args.emb_dim]),
            requires_grad=False,
        )

        self.ent_emb = nn.Embedding(self.args.num_ent, self.args.emb_dim)
        self.rel_emb = nn.Embedding(self.args.num_rel, self.args.emb_dim)
        self.transfer_matrix = nn.Embedding(
            self.args.num_rel, self.args.emb_dim * self.args.emb_dim
        )
        nn.init.uniform_(
            tensor=self.ent_emb.weight.data,
            a=-self.embedding_range.item(),
            b=self.embedding_range.item(),
        )
        nn.init.uniform_(
            tensor=self.rel_emb.weight.data,
            a=-self.embedding_range.item(),
            b=self.embedding_range.item(),
        )
        diag_matrix = torch.eye(self.args.emb_dim)
        diag_matrix = diag_matrix.flatten().repeat(self.args.num_rel, 1)
        self.transfer_matrix.weight.data = diag_matrix

        # nn.init.uniform_(tensor=self.transfer_matrix.weight.data, a=-self.embedding_range.item(), b=self.embedding_range.item())

    def score_func(self, head_emb, relation_emb, tail_emb, mode):
        """Calculating the score of triples.

        The formula for calculating the score is :math:`\gamma - \| M_{r} {e}_h + r_r - M_{r}e_t \|_{p}^2`

        """
        if self.norm_flag:
            head_emb = F.normalize(head_emb, 2, -1)
            relation_emb = F.normalize(relation_emb, 2, -1)
            tail_emb = F.normalize(tail_emb, 2, -1)
        if mode == "head-batch" or mode == "head_predict":
            score = head_emb + (relation_emb - tail_emb)
        else:
            score = (head_emb + relation_emb) - tail_emb
        score = self.margin.item() - torch.norm(score, p=1, dim=-1)
        return score

    def forward(self, triples, negs=None, mode="single"):
        """The functions used in the training phase, calculate triple score."""
        head_emb, relation_emb, tail_emb = self.tri2emb(triples, negs, mode)
        rel_transfer = self.transfer_matrix(triples[:, 1])  # shape:[bs, dim]
        head_emb = self._transfer(head_emb, rel_transfer, mode)
        tail_emb = self._transfer(tail_emb, rel_transfer, mode)
        score = self.score_func(head_emb, relation_emb, tail_emb, mode)

        return score

    def get_score(self, batch, mode):
        """The functions used in the testing phase, predict triple score."""
        triples = batch["positive_sample"]
        head_emb, relation_emb, tail_emb = self.tri2emb(triples, mode=mode)
        rel_transfer = self.transfer_matrix(triples[:, 1])  # shape:[bs, dim]
        head_emb = self._transfer(head_emb, rel_transfer, mode)
        tail_emb = self._transfer(tail_emb, rel_transfer, mode)
        score = self.score_func(head_emb, relation_emb, tail_emb, mode)

        return score

    def _transfer(self, emb, rel_transfer, mode):
        """Transfer entity embedding with relation-specific matrix.

        Args:
            emb: Entity embeddings, shape:[batch_size, emb_dim]
            rel_transfer: Relation-specific projection matrix, shape:[batch_size, emb_dim]
            mode: Choose head-predict or tail-predict, Defaults to 'single'.

        Returns:
            transfered entity emb: Shape:[batch_size, emb_dim]

        """
        rel_transfer = rel_transfer.view(-1, self.args.emb_dim, self.args.emb_dim)
        rel_transfer = rel_transfer.unsqueeze(dim=1)
        emb = emb.unsqueeze(dim=-2)
        emb = torch.matmul(emb, rel_transfer)
        return emb.squeeze(dim=-2)



import numpy as np
from torch.utils.data import Dataset
import torch
import os
from collections import defaultdict as ddict
from IPython import embed


class KGData(object):
    """Data preprocessing of kg data.

    Attributes:
        args: Some pre-set parameters, such as dataset path, etc. 
        ent2id: Encoding the entity in triples, type: dict.
        rel2id: Encoding the relation in triples, type: dict.
        id2ent: Decoding the entity in triples, type: dict.
        id2rel: Decoding the realtion in triples, type: dict.
        train_triples: Record the triples for training, type: list.
        valid_triples: Record the triples for validation, type: list.
        test_triples: Record the triples for testing, type: list.
        all_true_triples: Record all triples including train,valid and test, type: list.
        TrainTriples
        Relation2Tuple
        RelSub2Obj
        hr2t_train: Record the tail corresponding to the same head and relation, type: defaultdict(class:set).
        rt2h_train: Record the head corresponding to the same tail and relation, type: defaultdict(class:set).
        h2rt_train: Record the tail, relation corresponding to the same head, type: defaultdict(class:set).
        t2rh_train: Record the head, realtion corresponding to the same tail, type: defaultdict(class:set).
    """

    # TODO:把里面的函数再分一分，最基础的部分再初始化的使用调用，其他函数具体情况再调用
    def __init__(self, args):
        self.args = args

        #  基础部分
        self.ent2id = {}
        self.rel2id = {}
        # predictor需要
        self.id2ent = {}
        self.id2rel = {}
        # 存放三元组的id
        self.train_triples = []
        self.valid_triples = []
        self.test_triples = []
        self.all_true_triples = set()
        #  grounding 使用
        self.TrainTriples = {}
        self.Relation2Tuple = {}
        self.RelSub2Obj = {}

        self.hr2t_train = ddict(set)
        self.rt2h_train = ddict(set)
        self.h2rt_train = ddict(set)
        self.t2rh_train = ddict(set)
        self.get_id()
        self.get_triples_id()
        if args.use_weight:
            self.count = self.count_frequency(self.train_triples)

    def get_id(self):
        """Get entity/relation id, and entity/relation number.

        Update:
            self.ent2id: Entity to id.
            self.rel2id: Relation to id.
            self.id2ent: id to Entity.
            self.id2rel: id to Relation.
            self.args.num_ent: Entity number.
            self.args.num_rel: Relation number.
        """
        with open(os.path.join(self.args.data_path, "entities.dict")) as fin:
            for line in fin:
                eid, entity = line.strip().split("\t")
                self.ent2id[entity] = int(eid)
                self.id2ent[int(eid)] = entity

        with open(os.path.join(self.args.data_path, "relations.dict")) as fin:
            for line in fin:
                rid, relation = line.strip().split("\t")
                self.rel2id[relation] = int(rid)
                self.id2rel[int(rid)] = relation

        self.args.num_ent = len(self.ent2id)
        self.args.num_rel = len(self.rel2id)

    def get_triples_id(self):
        """Get triples id, save in the format of (h, r, t).

        Update:
            self.train_triples: Train dataset triples id.
            self.valid_triples: Valid dataset triples id.
            self.test_triples: Test dataset triples id.
        """
        
        with open(os.path.join(self.args.data_path, "train.txt")) as f:
            for line in f.readlines():
                h, r, t = line.strip().split()
                self.train_triples.append(
                    (self.ent2id[h], self.rel2id[r], self.ent2id[t])
                )
                
                tmp = str(self.ent2id[h]) + '\t' + str(self.rel2id[r]) + '\t' + str(self.ent2id[t])
                self.TrainTriples[tmp] = True

                iRelationID = self.rel2id[r]
                strValue = str(h) + "#" + str(t)
                if not iRelationID in self.Relation2Tuple:
                    tmpLst = []
                    tmpLst.append(strValue)
                    self.Relation2Tuple[iRelationID] = tmpLst
                else:
                    self.Relation2Tuple[iRelationID].append(strValue)

                iRelationID = self.rel2id[r]
                iSubjectID = self.ent2id[h]
                iObjectID = self.ent2id[t]
                tmpMap = {}
                tmpMap_in = {}
                if not iRelationID in self.RelSub2Obj:
                    if not iSubjectID in tmpMap:
                        tmpMap_in.clear()
                        tmpMap_in[iObjectID] = True
                        tmpMap[iSubjectID] = tmpMap_in
                    else:
                        tmpMap[iSubjectID][iObjectID] = True
                    self.RelSub2Obj[iRelationID] = tmpMap
                else:
                    tmpMap = self.RelSub2Obj[iRelationID]
                    if not iSubjectID in tmpMap:
                        tmpMap_in.clear()
                        tmpMap_in[iObjectID] = True
                        tmpMap[iSubjectID] = tmpMap_in
                    else:
                        tmpMap[iSubjectID][iObjectID] = True
                    self.RelSub2Obj[iRelationID] = tmpMap  # 是不是应该要加？

        with open(os.path.join(self.args.data_path, "valid.txt")) as f:
            for line in f.readlines():
                h, r, t = line.strip().split()
                self.valid_triples.append(
                    (self.ent2id[h], self.rel2id[r], self.ent2id[t])
                )

        with open(os.path.join(self.args.data_path, "test.txt")) as f:
            for line in f.readlines():
                h, r, t = line.strip().split()
                self.test_triples.append(
                    (self.ent2id[h], self.rel2id[r], self.ent2id[t])
                )

        self.all_true_triples = set(
            self.train_triples + self.valid_triples + self.test_triples
        )

    def get_hr2t_rt2h_from_train(self):
        """Get the set of hr2t and rt2h from train dataset, the data type is numpy.

        Update:
            self.hr2t_train: The set of hr2t.
            self.rt2h_train: The set of rt2h.
        """
        
        for h, r, t in self.train_triples:
            self.hr2t_train[(h, r)].add(t)
            self.rt2h_train[(r, t)].add(h)
        for h, r in self.hr2t_train:
            self.hr2t_train[(h, r)] = np.array(list(self.hr2t_train[(h, r)]))
        for r, t in self.rt2h_train:
            self.rt2h_train[(r, t)] = np.array(list(self.rt2h_train[(r, t)]))

    @staticmethod
    def count_frequency(triples, start=4):
        '''Get frequency of a partial triple like (head, relation) or (relation, tail).
        
        The frequency will be used for subsampling like word2vec.
        
        Args:
            triples: Sampled triples.
            start: Initial count number.

        Returns:
            count: Record the number of (head, relation).
        '''
        count = {}
        for head, relation, tail in triples:
            if (head, relation) not in count:
                count[(head, relation)] = start
            else:
                count[(head, relation)] += 1

            if (tail, -relation-1) not in count:
                count[(tail, -relation-1)] = start
            else:
                count[(tail, -relation-1)] += 1
        return count
        

    def get_h2rt_t2hr_from_train(self):
        """Get the set of h2rt and t2hr from train dataset, the data type is numpy.

        Update:
            self.h2rt_train: The set of h2rt.
            self.t2rh_train: The set of t2hr.
        """
        for h, r, t in self.train_triples:
            self.h2rt_train[h].add((r, t))
            self.t2rh_train[t].add((r, h))
        for h in self.h2rt_train:
            self.h2rt_train[h] = np.array(list(self.h2rt_train[h]))
        for t in self.t2rh_train:
            self.t2rh_train[t] = np.array(list(self.t2rh_train[t]))
        
    def get_hr_trian(self):
        '''Change the generation mode of batch.
        Merging triples which have same head and relation for 1vsN training mode.

        Returns:
            self.train_triples: The tuple(hr, t) list for training
        '''
        self.t_triples = self.train_triples 
        self.train_triples = [ (hr, list(t)) for (hr,t) in self.hr2t_train.items()]

class BaseSampler(KGData):
    """Traditional random sampling mode.
    """
    def __init__(self, args):
        super().__init__(args)
        self.get_hr2t_rt2h_from_train()

    def corrupt_head(self, t, r, num_max=1):
        """Negative sampling of head entities.

        Args:
            t: Tail entity in triple.
            r: Relation in triple.
            num_max: The maximum of negative samples generated 

        Returns:
            neg: The negative sample of head entity filtering out the positive head entity.
        """
        tmp = torch.randint(low=0, high=self.args.num_ent, size=(num_max,)).numpy()
        if not self.args.filter_flag:
            return tmp
        mask = np.in1d(tmp, self.rt2h_train[(r, t)], assume_unique=True, invert=True)
        neg = tmp[mask]
        return neg

    def corrupt_tail(self, h, r, num_max=1):
        """Negative sampling of tail entities.

        Args:
            h: Head entity in triple.
            r: Relation in triple.
            num_max: The maximum of negative samples generated 

        Returns:
            neg: The negative sample of tail entity filtering out the positive tail entity.
        """
        tmp = torch.randint(low=0, high=self.args.num_ent, size=(num_max,)).numpy()
        if not self.args.filter_flag:
            return tmp
        mask = np.in1d(tmp, self.hr2t_train[(h, r)], assume_unique=True, invert=True)
        neg = tmp[mask]
        return neg

    def head_batch(self, h, r, t, neg_size=None):
        """Negative sampling of head entities.

        Args:
            h: Head entity in triple
            t: Tail entity in triple.
            r: Relation in triple.
            neg_size: The size of negative samples.

        Returns:
            The negative sample of head entity. [neg_size]
        """
        neg_list = []
        neg_cur_size = 0
        while neg_cur_size < neg_size:
            neg_tmp = self.corrupt_head(t, r, num_max=(neg_size - neg_cur_size) * 2)
            neg_list.append(neg_tmp)
            neg_cur_size += len(neg_tmp)
        return np.concatenate(neg_list)[:neg_size]

    def tail_batch(self, h, r, t, neg_size=None):
        """Negative sampling of tail entities.

        Args:
            h: Head entity in triple
            t: Tail entity in triple.
            r: Relation in triple.
            neg_size: The size of negative samples.

        Returns:
            The negative sample of tail entity. [neg_size]
        """
        neg_list = []
        neg_cur_size = 0
        while neg_cur_size < neg_size:
            neg_tmp = self.corrupt_tail(h, r, num_max=(neg_size - neg_cur_size) * 2)
            neg_list.append(neg_tmp)
            neg_cur_size += len(neg_tmp)
        return np.concatenate(neg_list)[:neg_size]

    def get_train(self):
        return self.train_triples

    def get_valid(self):
        return self.valid_triples

    def get_test(self):
        return self.test_triples

    def get_all_true_triples(self):
        return self.all_true_triples


class RevSampler(KGData):
    """Adding reverse triples in traditional random sampling mode.

    For each triple (h, r, t), generate the reverse triple (t, r`, h).
    r` = r + num_rel.

    Attributes:
        hr2t_train: Record the tail corresponding to the same head and relation, type: defaultdict(class:set).
        rt2h_train: Record the head corresponding to the same tail and relation, type: defaultdict(class:set).
    """
    def __init__(self, args):
        super().__init__(args)
        self.hr2t_train = ddict(set)
        self.rt2h_train = ddict(set)
        self.add_reverse_relation()
        self.add_reverse_triples()
        self.get_hr2t_rt2h_from_train()

    def add_reverse_relation(self):
        """Get entity/relation/reverse relation id, and entity/relation number.

        Update:
            self.ent2id: Entity id.
            self.rel2id: Relation id.
            self.args.num_ent: Entity number.
            self.args.num_rel: Relation number.
        """
        
        with open(os.path.join(self.args.data_path, "relations.dict")) as fin:
            len_rel2id = len(self.rel2id)
            for line in fin:
                rid, relation = line.strip().split("\t")
                self.rel2id[relation + "_reverse"] = int(rid) + len_rel2id
                self.id2rel[int(rid) + len_rel2id] = relation + "_reverse"
        self.args.num_rel = len(self.rel2id)

    def add_reverse_triples(self):
        """Generate reverse triples (t, r`, h).

        Update:
            self.train_triples: Triples for training.
            self.valid_triples: Triples for validation.
            self.test_triples: Triples for testing.
            self.all_ture_triples: All triples including train, valid and test.
        """

        with open(os.path.join(self.args.data_path, "train.txt")) as f:
            for line in f.readlines():
                h, r, t = line.strip().split()
                self.train_triples.append(
                    (self.ent2id[t], self.rel2id[r + "_reverse"], self.ent2id[h])
                )

        with open(os.path.join(self.args.data_path, "valid.txt")) as f:
            for line in f.readlines():
                h, r, t = line.strip().split()
                self.valid_triples.append(
                    (self.ent2id[t], self.rel2id[r + "_reverse"], self.ent2id[h])
                )

        with open(os.path.join(self.args.data_path, "test.txt")) as f:
            for line in f.readlines():
                h, r, t = line.strip().split()
                self.test_triples.append(
                    (self.ent2id[t], self.rel2id[r + "_reverse"], self.ent2id[h])
                )

        self.all_true_triples = set(
            self.train_triples + self.valid_triples + self.test_triples
        )

    def get_train(self):
        return self.train_triples

    def get_valid(self):
        return self.valid_triples

    def get_test(self):
        return self.test_triples

    def get_all_true_triples(self):
        return self.all_true_triples    
    
    def corrupt_head(self, t, r, num_max=1):
        """Negative sampling of head entities.

        Args:
            t: Tail entity in triple.
            r: Relation in triple.
            num_max: The maximum of negative samples generated 

        Returns:
            neg: The negative sample of head entity filtering out the positive head entity.
        """
        tmp = torch.randint(low=0, high=self.args.num_ent, size=(num_max,)).numpy()
        if not self.args.filter_flag:
            return tmp
        mask = np.in1d(tmp, self.rt2h_train[(r, t)], assume_unique=True, invert=True)
        neg = tmp[mask]
        return neg

    def corrupt_tail(self, h, r, num_max=1):
        """Negative sampling of tail entities.

        Args:
            h: Head entity in triple.
            r: Relation in triple.
            num_max: The maximum of negative samples generated 

        Returns:
            neg: The negative sample of tail entity filtering out the positive tail entity.
        """
        tmp = torch.randint(low=0, high=self.args.num_ent, size=(num_max,)).numpy()
        if not self.args.filter_flag:
            return tmp
        mask = np.in1d(tmp, self.hr2t_train[(h, r)], assume_unique=True, invert=True)
        neg = tmp[mask]
        return neg

    def head_batch(self, h, r, t, neg_size=None):
        """Negative sampling of head entities.

        Args:
            h: Head entity in triple
            t: Tail entity in triple.
            r: Relation in triple.
            neg_size: The size of negative samples.

        Returns:
            The negative sample of head entity. [neg_size]
        """
        neg_list = []
        neg_cur_size = 0
        while neg_cur_size < neg_size:
            neg_tmp = self.corrupt_head(t, r, num_max=(neg_size - neg_cur_size) * 2)
            neg_list.append(neg_tmp)
            neg_cur_size += len(neg_tmp)
        return np.concatenate(neg_list)[:neg_size]

    def tail_batch(self, h, r, t, neg_size=None):
        """Negative sampling of tail entities.

        Args:
            h: Head entity in triple
            t: Tail entity in triple.
            r: Relation in triple.
            neg_size: The size of negative samples.

        Returns:
            The negative sample of tail entity. [neg_size]
        """
        neg_list = []
        neg_cur_size = 0
        while neg_cur_size < neg_size:
            neg_tmp = self.corrupt_tail(h, r, num_max=(neg_size - neg_cur_size) * 2)
            neg_list.append(neg_tmp)
            neg_cur_size += len(neg_tmp)
        return np.concatenate(neg_list)[:neg_size]


"""Base DataModule class."""
from pathlib import Path
from typing import Dict
import argparse
import os
from torch.utils.data import DataLoader
from .base_data_module import *
import pytorch_lightning as pl


class KGDataModule(BaseDataModule):
    """
    Base DataModule.
    Learn more at https://pytorch-lightning.readthedocs.io/en/stable/datamodules.html
    """

    def __init__(
        self, args: argparse.Namespace = None, train_sampler=None, test_sampler=None
    ) -> None:
        super().__init__(args)
        self.eval_bs = self.args.eval_bs
        self.num_workers = self.args.num_workers
        self.train_sampler = train_sampler
        self.test_sampler = test_sampler


    def get_data_config(self):
        """Return important settings of the dataset, which will be passed to instantiate models."""
        return {
            "num_training_steps": self.num_training_steps,
            "num_labels": self.num_labels,
        }

    def prepare_data(self):
        """
        Use this method to do things that might write to disk or that need to be done only from a single GPU in distributed settings (so don't set state `self.x = y`).
        """
        pass

    def setup(self, stage=None):
        """
        Split into train, val, test, and set dims.
        Should assign `torch Dataset` objects to self.data_train, self.data_val, and optionally self.data_test.
        """
        self.data_train = self.train_sampler.get_train()
        self.data_val   = self.train_sampler.get_valid()
        self.data_test  = self.train_sampler.get_test()

    def get_train_bs(self):
        """Get batch size for training.

        If the num_batches isn`t zero, it will divide data_train by num_batches to get batch size.
        And if user don`t give batch size and num_batches=0, it will raise ValueError.

        Returns:
            self.args.train_bs: The batch size for training.
        """
        if self.args.num_batches != 0:
            self.args.train_bs = len(self.data_train) // self.args.num_batches
        elif self.args.train_bs == 0:
            raise ValueError("train_bs or num_batches must specify one")
        return self.args.train_bs

    def train_dataloader(self):
        self.train_bs = self.get_train_bs()
        return DataLoader(
            self.data_train,
            shuffle=True,
            batch_size=self.train_bs,
            num_workers=self.num_workers,
            pin_memory=True,
            drop_last=True,
            collate_fn=self.train_sampler.sampling,
        )

    def val_dataloader(self):
        return DataLoader(
            self.data_val,
            shuffle=False,
            batch_size=self.eval_bs,
            num_workers=self.num_workers,
            pin_memory=True,
            collate_fn=self.test_sampler.sampling,
        )

    def test_dataloader(self):
        return DataLoader(
            self.data_test,
            shuffle=False,
            batch_size=self.eval_bs,
            num_workers=self.num_workers,
            pin_memory=True,
            collate_fn=self.test_sampler.sampling,
        )



import torch.nn as nn
import torch
import torch.nn.functional as F
from .model import Model
from IPython import embed


class TransH(Model):
    """`Knowledge Graph Embedding by Translating on Hyperplanes`_ (TransH), which apply the translation from head to tail entity in a
    relational-specific hyperplane in order to address its inability to model one-to-many, many-to-one, and many-to-many relations.

    Attributes:
        args: Model configuration parameters.
        epsilon: Calculate embedding_range.
        margin: Calculate embedding_range and loss.
        embedding_range: Uniform distribution range.
        ent_emb: Entity embedding, shape:[num_ent, emb_dim].
        rel_emb: Relation embedding, shape:[num_rel, emb_dim].
        norm_vector: Relation-specific projection matrix, shape:[num_rel, emb_dim]

    .. _Knowledge Graph Embedding by Translating on Hyperplanes: https://ojs.aaai.org/index.php/AAAI/article/view/8870
    """

    def __init__(self, args):
        super(TransH, self).__init__(args)
        self.args = args
        self.ent_emb = None
        self.rel_emb = None
        self.norm_flag = args.norm_flag
        self.init_emb()

    def init_emb(self):
        self.epsilon = 2.0
        self.margin = nn.Parameter(
            torch.Tensor([self.args.margin]), requires_grad=False
        )
        self.embedding_range = nn.Parameter(
            torch.Tensor([(self.margin.item() + self.epsilon) / self.args.emb_dim]),
            requires_grad=False,
        )

        self.ent_emb = nn.Embedding(self.args.num_ent, self.args.emb_dim)
        self.rel_emb = nn.Embedding(self.args.num_rel, self.args.emb_dim)
        self.norm_vector = nn.Embedding(self.args.num_rel, self.args.emb_dim)
        nn.init.uniform_(
            tensor=self.ent_emb.weight.data,
            a=-self.embedding_range.item(),
            b=self.embedding_range.item(),
        )
        nn.init.uniform_(
            tensor=self.rel_emb.weight.data,
            a=-self.embedding_range.item(),
            b=self.embedding_range.item(),
        )
        nn.init.uniform_(
            tensor=self.norm_vector.weight.data,
            a=-self.embedding_range.item(),
            b=self.embedding_range.item(),
        )

    def score_func(self, head_emb, relation_emb, tail_emb, mode):
        """Calculating the score of triples.

        The formula for calculating the score is :math:`\gamma - \|e'_{h,r} + d_r - e'_{t,r}\|_{p}^2`

        Args:
            head_emb: The head entity embedding.
            relation_emb: The relation embedding.
            tail_emb: The tail entity embedding.
            mode: Choose head-predict or tail-predict.

        Returns:
            score: The score of triples.
        """
        if self.norm_flag:
            head_emb = F.normalize(head_emb, 2, -1)
            relation_emb = F.normalize(relation_emb, 2, -1)
            tail_emb = F.normalize(tail_emb, 2, -1)
        if mode == "head-batch" or mode == "head_predict":
            score = head_emb + (relation_emb - tail_emb)
        else:
            score = (head_emb + relation_emb) - tail_emb
        score = self.margin.item() - torch.norm(score, p=1, dim=-1)
        return score

    def forward(self, triples, negs=None, mode="single"):
        """The functions used in the training phase, same as TransE"""

        head_emb, relation_emb, tail_emb = self.tri2emb(triples, negs, mode)
        norm_vector = self.norm_vector(triples[:, 1]).unsqueeze(
            dim=1
        )  # shape:[bs, 1, dim]
        head_emb = self._transfer(head_emb, norm_vector)
        tail_emb = self._transfer(tail_emb, norm_vector)
        score = self.score_func(head_emb, relation_emb, tail_emb, mode)

        return score

    def get_score(self, batch, mode):
        """The functions used in the testing phase, same as TransE"""

        triples = batch["positive_sample"]
        head_emb, relation_emb, tail_emb = self.tri2emb(triples, mode=mode)
        norm_vector = self.norm_vector(triples[:, 1]).unsqueeze(
            dim=1
        )  # shape:[bs, 1, dim]
        head_emb = self._transfer(head_emb, norm_vector)
        tail_emb = self._transfer(tail_emb, norm_vector)
        score = self.score_func(head_emb, relation_emb, tail_emb, mode)
        return score

    def _transfer(self, emb, norm_vector):
        """Projecting entity embeddings onto the relation-specific hyperplane

        The formula for Projecting entity embeddings is :math:`e'_{r} = e - w_r^\Top e w_r`

        Args:
            emb: Entity embeddings, shape:[batch_size, emb_dim]
            norm_vector: Relation-specific projection matrix, shape:[num_rel, emb_dim]

        Returns:
            projected entity emb: Shape:[batch_size, emb_dim]

        """
        if self.norm_flag:
            norm_vector = F.normalize(norm_vector, p=2, dim=-1)
        return emb - torch.sum(emb * norm_vector, -1, True) * norm_vector



import torch.nn as nn
import torch
import torch.nn.functional as F
import math
from .model import Model
from IPython import embed


class SimplE(Model):
    """`SimplE Embedding for Link Prediction in Knowledge Graphs`_ (SimpleE), which presents a simple enhancement of CP (which we call SimplE) to allow the two embeddings of each entity to be learned dependently.

    Attributes:
        args: Model configuration parameters.
        epsilon: Calculate embedding_range.
        margin: Calculate embedding_range and loss.
        embedding_range: Uniform distribution range.
        ent_h_emb: Entity embedding, shape:[num_ent, emb_dim].
        ent_t_emb: Entity embedding, shape:[num_ent, emb_dim].
        rel_emb: Relation_embedding, shape:[num_rel, emb_dim].
        rel_inv_emb: Inverse Relation_embedding, shape:[num_rel, emb_dim].

    .. _SimplE Embedding for Link Prediction in Knowledge Graphs: http://papers.neurips.cc/paper/7682-simple-embedding-for-link-prediction-in-knowledge-graphs.pdf
    """
    def __init__(self, args):
        super(SimplE, self).__init__(args)
        self.args = args
        self.ent_emb = None
        self.rel_emb = None
        self.init_emb()

    def init_emb(self):
        """Initialize the entity and relation embeddings in the form of a uniform distribution."""
        self.ent_h_emb = nn.Embedding(self.args.num_ent, self.args.emb_dim)
        self.ent_t_emb = nn.Embedding(self.args.num_ent, self.args.emb_dim)
        self.rel_emb = nn.Embedding(self.args.num_rel, self.args.emb_dim)
        self.rel_inv_emb = nn.Embedding(self.args.num_rel, self.args.emb_dim)
        sqrt_size = 6.0 / math.sqrt(self.args.emb_dim)
        nn.init.uniform_(tensor=self.ent_h_emb.weight.data, a=-sqrt_size, b=sqrt_size)
        nn.init.uniform_(tensor=self.ent_t_emb.weight.data, a=-sqrt_size, b=sqrt_size)
        nn.init.uniform_(tensor=self.rel_emb.weight.data, a=-sqrt_size, b=sqrt_size)
        nn.init.uniform_(tensor=self.rel_inv_emb.weight.data, a=-sqrt_size, b=sqrt_size)

    def score_func(self, hh_emb, rel_emb, tt_emb, ht_emb, rel_inv_emb, th_emb):
        """Calculating the score of triples.

        Args:
            hh_emb: The head entity embedding on embedding h.
            rel_emb: The relation embedding.
            tt_emb: The tail entity embedding on embedding t.
            ht_emb: The tail entity embedding on embedding h.
            rel_inv_emb: The tail entity embedding.
            th_emb: The head entity embedding on embedding t.

        Returns:
            score: The score of triples.
        """
        # return -(torch.sum(head_emb * relation_emb * tail_emb, -1) + \
        #     torch.sum(head_emb * rel_inv_emb * tail_emb, -1))/2
        scores1 = torch.sum(hh_emb * rel_emb * tt_emb, dim=-1)
        scores2 = torch.sum(ht_emb * rel_inv_emb * th_emb, dim=-1)
        return torch.clamp((scores1 + scores2) / 2, -20, 20)

    def l2_loss(self):
        return (self.ent_h_emb.weight.norm(p = 2) ** 2 + \
            self.ent_t_emb.weight.norm(p = 2) ** 2 + \
            self.rel_emb.weight.norm(p = 2) ** 2 + \
            self.rel_inv_emb.weight.norm(p = 2) ** 2)
    
    def forward(self, triples, negs=None, mode='single'):
        """The functions used in the training phase

        Args:
            triples: The triples ids, as (h, r, t), shape:[batch_size, 3].
            negs: Negative samples, defaults to None.
            mode: Choose head-predict or tail-predict, Defaults to 'single'.

        Returns:
            score: The score of triples.
        """
        rel_emb, rel_inv_emb, hh_emb, th_emb, ht_emb, tt_emb = self.get_emb(triples, negs, mode)
        return self.score_func(hh_emb, rel_emb, tt_emb, ht_emb, rel_inv_emb, th_emb)
    
    def get_score(self, batch, mode):
        """The functions used in the testing phase

        Args:
            batch: A batch of data.
            mode: Choose head-predict or tail-predict.

        Returns:
            score: The score of triples.
        """
        triples = batch["positive_sample"]
        rel_emb, rel_inv_emb, hh_emb, th_emb, ht_emb, tt_emb = self.get_emb(triples, mode=mode)
        return self.score_func(hh_emb, rel_emb, tt_emb, ht_emb, rel_inv_emb, th_emb)
    
    def get_emb(self, triples, negs=None, mode='single'):
        if mode == 'single':
            rel_emb = self.rel_emb(triples[:, 1]).unsqueeze(1)  # [bs, 1, dim]
            rel_inv_emb = self.rel_inv_emb(triples[:, 1]).unsqueeze(1)
            hh_emb = self.ent_h_emb(triples[:, 0]).unsqueeze(1)  # [bs, 1, dim]
            th_emb = self.ent_t_emb(triples[:, 0]).unsqueeze(1)  # [bs, 1, dim]
            ht_emb = self.ent_h_emb(triples[:, 2]).unsqueeze(1)  # [bs, 1, dim]
            tt_emb = self.ent_t_emb(triples[:, 2]).unsqueeze(1)  # [bs, 1, dim]
        elif mode == 'head-batch' or mode == "head_predict":
            if negs is None:  # 说明这个时候是在evluation，所以需要直接用所有的entity embedding
                hh_emb = self.ent_h_emb.weight.data.unsqueeze(0)  # [1, num_ent, dim]
                th_emb = self.ent_t_emb.weight.data.unsqueeze(0)  # [1, num_ent, dim]
            else:
                hh_emb = self.ent_h_emb(negs)  # [bs, num_neg, dim]
                th_emb = self.ent_t_emb(negs)  # [bs, num_neg, dim]

            rel_emb = self.rel_emb(triples[:, 1]).unsqueeze(1)  # [bs, 1, dim]
            rel_inv_emb = self.rel_inv_emb(triples[:, 1]).unsqueeze(1)  # [bs, 1, dim]
            ht_emb = self.ent_h_emb(triples[:, 2]).unsqueeze(1)  # [bs, 1, dim]
            tt_emb = self.ent_t_emb(triples[:, 2]).unsqueeze(1)  # [bs, 1, dim]
        elif mode == 'tail-batch' or mode == "tail_predict":
            if negs is None:  # 说明这个时候是在evluation，所以需要直接用所有的entity embedding
                ht_emb = self.ent_h_emb.weight.data.unsqueeze(0)  # [1, num_ent, dim]
                tt_emb = self.ent_t_emb.weight.data.unsqueeze(0)  # [1, num_ent, dim]
            else:
                ht_emb = self.ent_h_emb(negs)  # [bs, num_neg, dim]
                tt_emb = self.ent_t_emb(negs)  # [bs, num_neg, dim]

            rel_emb = self.rel_emb(triples[:, 1]).unsqueeze(1)  # [bs, 1, dim]
            rel_inv_emb = self.rel_inv_emb(triples[:, 1]).unsqueeze(1)  # [bs, 1, dim]
            hh_emb = self.ent_h_emb(triples[:, 0]).unsqueeze(1)  # [bs, 1, dim]
            th_emb = self.ent_t_emb(triples[:, 0]).unsqueeze(1)  # [bs, 1, dim]

        return rel_emb, rel_inv_emb, hh_emb, th_emb, ht_emb, tt_emb




import torch
import torch.nn as nn
import math
from torch.autograd import Variable
from IPython import embed
class RugE_Loss(nn.Module):
    def __init__(self,args, model):
        super(RugE_Loss, self).__init__()
        self.args = args
        self.model = model

    def forward(self, pos_score, neg_score, rule, confidence, triple_num, pos_len):
        entroy = nn.BCELoss()

        # 这段代码写的太简陋了 先跑通再说
        pos_label = torch.ones([pos_len, 1])
        neg_label = torch.zeros([pos_len, self.args.num_neg])
        one = torch.ones([1])
        zero = torch.zeros([1])
        pos_label = Variable(pos_label).to(self.args.gpu, dtype=torch.float)
        neg_label = Variable(neg_label).to(self.args.gpu, dtype=torch.float)
        one = Variable(one).to(self.args.gpu, dtype=torch.float)
        zero = Variable(zero).to(self.args.gpu, dtype=torch.float)

        sigmoid_neg = torch.sigmoid(neg_score)
        sigmoid_pos = torch.sigmoid(pos_score)

        postive_loss = entroy(sigmoid_pos, pos_label)
        negative_loss = entroy(sigmoid_neg, neg_label)

        pi_gradient = dict()
        # 感觉应该放在这个大函数的外面，不然每次被清空也没什么用
        sigmoid_value = dict()
        # 在计算每个grounding rule中的unlable的三元组对应的类似gradient
        for i in range(len(rule[0])):
            if triple_num[i] == 2:
                p1_rule = rule[0][i]
                unlabel_rule = rule[1][i]

                if p1_rule not in sigmoid_value:
                    p1_rule_score = self.model(p1_rule.unsqueeze(0))
                    sigmoid_rule = torch.sigmoid(p1_rule_score)
                    sigmoid_value[p1_rule] = sigmoid_rule
                else:
                    sigmoid_rule = sigmoid_value[p1_rule]

                if unlabel_rule not in pi_gradient:
                    pi_gradient[unlabel_rule] = self.args.slackness_penalty * confidence[i] * sigmoid_rule
                else:
                    pi_gradient[unlabel_rule] += self.args.slackness_penalty * confidence[i] * sigmoid_rule

            elif triple_num[i] == 3:
                p1_rule = rule[0][i]
                p2_rule = rule[1][i]
                unlabel_rule = rule[2][i]

                if p1_rule not in sigmoid_value:
                    p1_rule_score = self.model(p1_rule.unsqueeze(0))
                    sigmoid_rule = torch.sigmoid(p1_rule_score)
                    sigmoid_value[p1_rule] = sigmoid_rule
                else:
                    sigmoid_rule = sigmoid_value[p1_rule]

                if p2_rule not in sigmoid_value:
                    p2_rule_score = self.model(p2_rule.unsqueeze(0))
                    sigmoid_rule2 = torch.sigmoid(p2_rule_score)
                    sigmoid_value[p2_rule] = sigmoid_rule
                else:
                    sigmoid_rule2 = sigmoid_value[p2_rule]

                if unlabel_rule not in pi_gradient:
                    pi_gradient[unlabel_rule] = self.args.slackness_penalty * confidence[i] * sigmoid_rule * sigmoid_rule2
                else:
                    pi_gradient[unlabel_rule] += self.args.slackness_penalty * confidence[i] * sigmoid_rule * sigmoid_rule2

        unlabel_loss = 0.
        unlabel_triples = []
        gradient = []
        # 对于pi_gradient中的每个三元组（不重复）的 根据公式计算s函数
        for unlabel_triple in pi_gradient.keys():
            unlabel_triples.append(unlabel_triple.cpu().numpy())
            gradient.append(pi_gradient[unlabel_triple].cpu().detach().numpy())
        unlabel_triples = torch.tensor(unlabel_triples).to(self.args.gpu)
        gradient = torch.tensor(gradient).to(self.args.gpu).view(-1, 1)
        unlabel_triple_score = self.model(unlabel_triples)
        unlabel_triple_score = torch.sigmoid(unlabel_triple_score)
        unlabel_scores = []
        for i in range(0, len(gradient)):
            unlabel_score = (torch.min(torch.max(unlabel_triple_score[i] + gradient[i], zero), one)).cpu().detach().numpy()
            unlabel_scores.append(unlabel_score[0])
        unlabel_scores = torch.tensor(unlabel_scores).to(self.args.gpu)
        unlabel_scores = unlabel_scores.unsqueeze(1)
        unlabel_loss = entroy(unlabel_triple_score, unlabel_scores)
        # for unlabel_triple in pi_gradient.keys():
        #     unlabelrule_score = model(unlabel_triple.unsqueeze(0))
        #     sigmoid_unlabelrule = torch.sigmoid(unlabelrule_score)
        #     unlabel_score = torch.min(torch.max(sigmoid_unlabelrule + args.slackness_penalty * pi_gradient[unlabel_triple], zero), one)
        #     loss_part = entroy(sigmoid_unlabelrule, unlabel_score.to(args.gpu).detach())
        #     unlabel_loss = unlabel_loss + loss_part
        # 所有的grounding的unlbeled的两个值sigmoid和s函数都存在list中，需要转成tensor，然后一起计算loss

        loss = postive_loss + negative_loss + unlabel_loss

        if self.args.weight_decay != 0.0:
            #Use L2 regularization for ComplEx_NNE_AER
            ent_emb_all = self.model.ent_emb(torch.arange(self.args.num_ent).to(self.args.gpu))
            rel_emb_all = self.model.rel_emb(torch.arange(self.args.num_rel).to(self.args.gpu))
            regularization = self.args.weight_decay * (
                ent_emb_all.norm(p = 2)**2 + rel_emb_all.norm(p=2)**2
            )

        # print(postive_loss)
        # print(negative_loss)
        # print(unlabel_loss)
        loss += regularization
        return loss









import importlib
from IPython import embed
import os
import time
import yaml

def import_class(module_and_class_name: str) -> type:
    """Import class from a module, e.g. 'model.TransE'"""
    module_name, class_name = module_and_class_name.rsplit(".", 1)
    module = importlib.import_module(module_name)
    class_ = getattr(module, class_name)
    return class_


def save_config(args):
    args.save_config = False  #防止和get_config冲突，导致把加载的config又保存了一遍
    if not os.path.exists("config"):
        os.mkdir("config")
    config_file_name = time.strftime("%H:%M:%S", time.localtime()) + ".yaml"
    day_name = time.strftime("%Y-%m-%d")
    if not os.path.exists(os.path.join("config", day_name)):
        os.makedirs(os.path.join("config", day_name))
    config = vars(args)
    with open(os.path.join(os.path.join("config", day_name), config_file_name), "w") as file:
        file.write(yaml.dump(config))

def load_config(args, config_path):
    with open(config_path, "r") as f:
        config = yaml.safe_load(f)
        args.__dict__.update(config)
    return args



from logging import debug
import pytorch_lightning as pl
import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
import os
import json
from collections import defaultdict as ddict
from IPython import embed
from .BaseLitModel import BaseLitModel
from neuralkg.eval_task import *
from IPython import embed

from functools import partial

class CrossELitModel(BaseLitModel):
    def __init__(self, model, args):
        super().__init__(model, args)
        
    def forward(self, x):
        return self.model(x)
    
    def training_step(self, batch, batch_idx):

        sample = batch["sample"]
        hr_label  = batch["hr_label"]
        tr_label  = batch["tr_label"]
        # sample_id = batch["sample_id"]
        # sample_score = self.model(sample, sample_id)
        hr_score, tr_score = self.model(sample)
        hr_loss = self.loss(hr_score, hr_label)
        tr_loss = self.loss(tr_score, tr_label)
        loss = hr_loss + tr_loss
        regularize_loss = self.args.weight_decay * self.model.regularize_loss(1)
        loss += regularize_loss
        self.log("Train|loss", loss,  on_step=False, on_epoch=True)
        return loss
    
    def validation_step(self, batch, batch_idx):
        # pos_triple, tail_label, head_label = batch
        results = dict()
        ranks = link_predict(batch, self.model, predicion='all')
        results["count"] = torch.numel(ranks)
        results["mrr"] = torch.sum(1.0 / ranks).item()
        for k in self.args.calc_hits:
            results['hits@{}'.format(k)] = torch.numel(ranks[ranks <= k])
        return results
    
    def validation_epoch_end(self, results) -> None:
        outputs = self.collect_results(results, "Eval")
        # self.log("Eval|mrr", outputs["Eval|mrr"], on_epoch=True)
        self.log_dict(outputs, prog_bar=True, on_epoch=True)

    def test_step(self, batch, batch_idx):
        results = dict()
        ranks = link_predict(batch, self.model, predicion='all')
        results["count"] = torch.numel(ranks)
        results["mrr"] = torch.sum(1.0 / ranks).item()
        for k in self.args.calc_hits:
            results['hits@{}'.format(k)] = torch.numel(ranks[ranks <= k])
        return results
    
    def test_epoch_end(self, results) -> None:
        outputs = self.collect_results(results, "Test")
        self.log_dict(outputs, prog_bar=True, on_epoch=True)

    '''这里设置优化器和lr_scheduler'''
    def configure_optimizers(self):
        milestones = int(self.args.max_epochs / 2)
        # optimizer = self.optimizer_class(self.model.parameters(), lr=self.args.lr, weight_decay = self.args.weight_decay)
        optimizer = self.optimizer_class(self.model.parameters(), lr=self.args.lr)
        StepLR = torch.optim.lr_scheduler.MultiStepLR(optimizer, milestones=[milestones], gamma=0.1)
        optim_dict = {'optimizer': optimizer, 'lr_scheduler': StepLR}
        return optim_dict



import torch
import torch.nn.functional as F
import torch.nn as nn
from IPython import embed


class CrossE_Loss(nn.Module):
    def __init__(self, args, model):
        super(CrossE_Loss, self).__init__()
        self.args = args
        self.model = model

    def forward(self, score, label):
        return -torch.sum(torch.log(torch.clamp(score, 1e-10, 1.0)) * torch.clamp(label, 0.0, 1.0) + \
            torch.log(torch.clamp(1-score, 1e-10, 1.0)) * torch.clamp(-label, 0.0, 1.0))



# -*- coding: utf-8 -*-
import argparse
import os
import yaml
import pytorch_lightning as pl
from neuralkg import lit_model
from neuralkg import data
def setup_parser():
    """Set up Python's ArgumentParser with data, model, trainer, and other arguments."""
    parser = argparse.ArgumentParser(add_help=False)

    # Add Trainer specific arguments, such as --max_epochs, --gpus, --precision
    trainer_parser = pl.Trainer.add_argparse_args(parser)
    trainer_parser._action_groups[1].title = "Trainer Args"  # pylint: disable=protected-access
    parser = argparse.ArgumentParser(add_help=False, parents=[trainer_parser])

    # Basic arguments
    parser.add_argument('--name', default='ComplEX_FB15K', type=str)
    parser.add_argument('--num_epoch', default=1000, type=int)
    parser.add_argument("--use_wandb", default=True, action='store_false')
    parser.add_argument('--norm_flag', default=False, action='store_true')
    parser.add_argument('--test_only', default=False, action='store_true')
    parser.add_argument('--filter_flag', default=True, action='store_false')
    parser.add_argument('--save_config', default=False, action='store_true')
    parser.add_argument('--load_config', default=False, action='store_true')
    parser.add_argument('--negative_adversarial_sampling','-adv', default=True, action='store_false')
    parser.add_argument('--use_weight', default=False, action='store_true')
    parser.add_argument("--seed", default=666, type=int)
    parser.add_argument("--litmodel_name", default="KGELitModel", type=str)
    parser.add_argument('--model_name', default="TransE", type=str)
    parser.add_argument('--loss_name', default="Adv_Loss", type=str)
    parser.add_argument('--dataset_name', default="FB15K237", type=str)
    parser.add_argument('--optim_name', default="Adam", type=str)
    parser.add_argument('--margin', default=12.0, type=float)
    parser.add_argument('--checkpoint_dir', default="", type=str)
    parser.add_argument('--regularization', '-r', default=0.0, type=float) #后续尝试使用optim的正则化方法
    parser.add_argument('--mu', default=10, type=float, help='penalty coefficient for ComplEx_NNE')
    parser.add_argument('--emb_shape', default=20, type=int, help='Only on ConvE,The first dimension of the reshaped 2D embedding')
    parser.add_argument('--inp_drop', default=0.2, type=float, help='only on ConvE,Dropout for the input embeddings')
    parser.add_argument('--hid_drop', default=0.3, type=float, help='only on ConvE,Dropout for the hidden layer')
    parser.add_argument('--fet_drop', default=0.2, type=float, help='only on ConvE,Dropout for the convolutional features')
    parser.add_argument('--hid_size', default=9728, type=int, help='only on ConvE,The side of the hidden layer. The required size changes with the size of the embeddings.')
    parser.add_argument('--smoothing', default=0.1, type=float, help='only on ConvE,Make the label smooth')
    parser.add_argument('--dropout', default=0.5, type=float, help='only on CrossE,for Dropout')
    parser.add_argument('--neg_weight', default=50, type=int, help='only on CrossE, make up label')
    parser.add_argument('--slackness_penalty', default=0.01, type=float)
    parser.add_argument('--check_per_epoch', default=5, type=int)
    parser.add_argument('--early_stop_patience', default=5, type=int)
    parser.add_argument('--emb_dim', default=200, type=int)
    parser.add_argument('--out_dim', default=200, type=int)
    parser.add_argument('--num_neg', default=10, type=int)
    parser.add_argument('--num_ent', default=None, type=int)
    parser.add_argument('--num_rel', default=None, type=int)
    parser.add_argument('--freq_init', default=4, type=int)
    parser.add_argument('--eval_task', default="link_prediction", type=str)
    parser.add_argument('--adv_temp', default=1.0, type=float)
    parser.add_argument('--data_class', default="KGDataModule", type=str)
    parser.add_argument("--train_sampler_class",default="UniSampler",type=str)
    parser.add_argument("--test_sampler_class",default="TestSampler",type=str)
    parser.add_argument("--decoder_model", default=None, type=str)
    parser.add_argument("--opn", default='corr',type=str, help="only on CompGCN, choose Composition Operation")
    
    parser.add_argument('--gpu', default='cuda:0', type=str)
    parser.add_argument('--shuffle', default=True, action='store_false')
    parser.add_argument('--bern_flag', default=False, action='store_true')
    parser.add_argument('--config_path', default="", type=str)

    
    # parser only for IterE
    parser.add_argument("--max_entialments", default=2000, type=int, help="in IterE.py")
    parser.add_argument("--axiom_types", default=10, type=int, help="in IterE.py")
    parser.add_argument("--select_probability", default=0.8, type=float, help="in IterE.py")
    parser.add_argument("--axiom_weight", default=1.0, type=float, help="in IterE.py")
    parser.add_argument("--inject_triple_percent", default=1.0, type=float, help="in IterE.py")
    parser.add_argument("--update_axiom_per",default=2, type=int, help='in IterELitModel.py')
    
    parser.add_argument("--calc_hits", default=[1,3,10], type=lambda s: [int(item) for item in s.split(',')], help='calc hits list')
    parser.add_argument("--dis_order", default=2, type=int, help="the distance order of score")
    # Get data, model, and LitModel specific arguments
    lit_model_group = parser.add_argument_group("LitModel Args")
    lit_model.BaseLitModel.add_to_argparse(lit_model_group)

    data_group = parser.add_argument_group("Data Args")
    data.BaseDataModule.add_to_argparse(data_group)


    parser.add_argument("--help", "-h", action="help")
    args = parser.parse_args()
    
    return args



from .BaseLitModel import BaseLitModel
from .KGELitModel import KGELitModel
from .ConvELitModel import ConvELitModel
from .RGCNLitModel import RGCNLitModel
from .KBATLitModel import KBATLitModel
from .CompGCNLitModel import CompGCNLitModel
from .CrossELitModel import CrossELitModel
from .XTransELitModel import XTransELitModel
from .IterELitModel import IterELitModel
from .RugELitModel import RugELitModel



#!/usr/bin/env python
# coding: utf-8

import setuptools
import os

requirement_path = os.path.realpath("requirements.txt")
if os.path.isfile(requirement_path):
    with open(requirement_path) as f:
        install_requires = f.read().splitlines()

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='neuralkg',
    version='1.0.3',
    author='ZJUKG',
    author_email='xnchen2020@zju.edu.cn',
    url='https://github.com/zjukg/NeuralKG',
    description='An Open Source Library for Diverse Representation Learning of Knowledge Graphs',
    package_dir={"": "neuralkg"},
    packages=setuptools.find_packages("neuralkg"),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=install_requires
)


import dgl
import torch
import torch.nn as nn
import torch.nn.functional as F
from dgl.nn.pytorch import RelGraphConv
from neuralkg.model import DistMult


class RGCN(nn.Module):
    """`Modeling Relational Data with Graph Convolutional Networks`_ (RGCN), which use GCN framework to model relation data.

    Attributes:
        args: Model configuration parameters.
    
    .. _Modeling Relational Data with Graph Convolutional Networks: https://arxiv.org/pdf/1703.06103.pdf
    """

    def __init__(self, args):
        super(RGCN, self).__init__()
        self.args = args
        self.ent_emb = None
        self.rel_emb = None 
        self.RGCN = None 
        self.build_model()

    def build_model(self):
        """Initialize the RGCN model and embeddings 

        Args:
            ent_emb: Entity embedding, shape:[num_ent, emb_dim].
            rel_emb: Relation_embedding, shape:[num_rel, emb_dim].
            RGCN: the relation graph convolution model.
        """
        self.ent_emb = nn.Embedding(self.args.num_ent,self.args.emb_dim)

        self.rel_emb = nn.Parameter(torch.Tensor(self.args.num_rel, self.args.emb_dim))

        nn.init.xavier_uniform_(self.rel_emb, gain=nn.init.calculate_gain('relu'))

        self.RGCN = nn.ModuleList()
        for idx in range(self.args.num_layers):
            RGCN_idx = self.build_hidden_layer(idx)
            self.RGCN.append(RGCN_idx)
        
    def forward(self, graph, ent, rel, norm, triples, mode='single'):
        
        """The functions used in the training and testing phase

        Args:
            graph: The knowledge graph recorded in dgl.graph()
            ent: The entitiy ids sampled in triples
            rel: The relation ids sampled in triples
            norm: The edge norm in graph 
            triples: The triples ids, as (h, r, t), shape:[batch_size, 3].
            mode: Choose head-predict or tail-predict, Defaults to 'single'.

        Returns:
            score: The score of triples.
        """
        
        embedding = self.ent_emb(ent.squeeze())
        for layer in self.RGCN:
            embedding = layer(graph, embedding, rel, norm)
        head_emb, rela_emb, tail_emb = self.tri2emb(embedding, triples, mode)
        score = DistMult.score_func(self,head_emb, rela_emb, tail_emb, mode)

        return score

    def tri2emb(self, embedding, triples, mode="single"):
        
        """Get embedding of triples.
        
        This function get the embeddings of head, relation, and tail
        respectively. each embedding has three dimensions.

        Args:
            embedding(tensor): This embedding save the entity embeddings.            
            triples (tensor): This tensor save triples id, which dimension is 
                [triples number, 3].
            mode (str, optional): This arg indicates that the negative entity 
                will replace the head or tail entity. when it is 'single', it 
                means that entity will not be replaced. Defaults to 'single'.

        Returns:
            head_emb: Head entity embedding.
            rela_emb: Relation embedding.
            tail_emb: Tail entity embedding.
        """

        rela_emb = self.rel_emb[triples[:, 1]].unsqueeze(1)  # [bs, 1, dim]
        head_emb = embedding[triples[:, 0]].unsqueeze(1)  # [bs, 1, dim] 
        tail_emb = embedding[triples[:, 2]].unsqueeze(1)  # [bs, 1, dim]

        if mode == "head-batch":
            head_emb = embedding.unsqueeze(0)  # [1, num_ent, dim]

        elif mode == "tail-batch":
            tail_emb = embedding.unsqueeze(0)  # [1, num_ent, dim]

        return head_emb, rela_emb, tail_emb

    def build_hidden_layer(self, idx):
        """The functions used to initialize the RGCN model

        Args:
            idx: it`s used to identify rgcn layers. The last rgcn layer should use 
            relu as activation function.

        Returns:
            the relation graph convolution layer
        """
        act = F.relu if idx < self.args.num_layers - 1 else None
        return RelGraphConv(self.args.emb_dim, self.args.emb_dim, self.args.num_rel, "bdd",
                    num_bases=100, activation=act, self_loop=True,dropout=0.2 )



import torch
import torch.nn.functional as F
import torch.nn as nn
from IPython import embed

class Margin_Loss(nn.Module):
    def __init__(self, args, model):
        super(Margin_Loss, self).__init__()
        self.args = args
        self.model = model

    def forward(self, pos_score, neg_score):
        neg_score = F.logsigmoid(-neg_score)  #shape:[bs]
        pos_score = F.logsigmoid(pos_score) #shape:[bs, 1]
        positive_sample_loss = - pos_score.mean()
        negative_sample_loss = - neg_score.mean()
        loss = (positive_sample_loss + negative_sample_loss) / 2

        if self.args.model_name == "XTransE":
            regularization = self.args.regularization * (
                self.model.ent_emb.weight.norm(p = 1) + \
                self.model.rel_emb.weight.norm(p = 1)
            )

        if self.args.model_name == 'ComplEx' or self.args.model_name == 'DistMult':
            #Use L3 regularization for ComplEx and DistMult
            regularization = self.args.regularization * (
                self.model.ent_emb.weight.norm(p = 3)**3 + \
                self.model.rel_emb.weight.norm(p = 3)**3
            )
            loss = loss + regularization
            
        return loss


import torch
import os
from IPython import embed


def link_predict(batch, model, predicion="all"):
    """The evaluate task is predicting the head entity or tail entity in incomplete triples.
        
    Args:
        batch: The batch of the triples for validation or test.
        model: The KG model for training.
        predicion: mode of link prediction.

    Returns:
        ranks: The rank of the triple to be predicted.
    """
    if predicion == "all":
        tail_ranks = tail_predict(batch, model)
        head_ranks = head_predict(batch, model)
        ranks = torch.cat([tail_ranks, head_ranks])
    elif predicion == "head":
        ranks = head_predict(batch, model)
    elif predicion == "tail":
        ranks = tail_predict(batch, model)

    return ranks.float()


def head_predict(batch, model):
    """Getting head entity ranks.

    Args:
        batch: The batch of the triples for validation or test
        model: The KG model for training.

    Returns:
        tensor: The rank of the head entity to be predicted, dim [batch_size]
    """
    pos_triple = batch["positive_sample"]
    idx = pos_triple[:, 0]
    label = batch["head_label"]
    pred_score = model.get_score(batch, "head_predict")
    return calc_ranks(idx, label, pred_score)


def tail_predict(batch, model):
    """Getting tail entity ranks.

    Args:
        batch: The batch of the triples for validation or test
        model: The KG model for training.

    Returns:
        tensor: The rank of the tail entity to be predicted, dim [batch_size]
    """
    pos_triple = batch["positive_sample"]
    idx = pos_triple[:, 2]
    label = batch["tail_label"]
    pred_score = model.get_score(batch, "tail_predict")
    return calc_ranks(idx, label, pred_score)


def calc_ranks(idx, label, pred_score):
    """Calculating triples score ranks.

    Args:
        idx ([type]): The id of the entity to be predicted.
        label ([type]): The id of existing triples, to calc filtered results.
        pred_score ([type]): The score of the triple predicted by the model.

    Returns:
        ranks: The rank of the triple to be predicted, dim [batch_size].
    """

    b_range = torch.arange(pred_score.size()[0])
    target_pred = pred_score[b_range, idx]
    pred_score = torch.where(label.byte(), -torch.ones_like(pred_score) * 10000000, pred_score)
    pred_score[b_range, idx] = target_pred

    ranks = (
        1
        + torch.argsort(
            torch.argsort(pred_score, dim=1, descending=True), dim=1, descending=False
        )[b_range, idx]
    )
    return ranks


import torch
import torch.nn as nn
from IPython import embed
from neuralkg.data import KGData


class ComplEx_NNE_AER_Loss(nn.Module):
    def __init__(self, args, model):
        super(ComplEx_NNE_AER_Loss, self).__init__()
        self.args = args
        self.model = model
        self.rule_p, self.rule_q = model.rule
        self.confidence = model.conf
        
    def forward(self, pos_score, neg_score):
        logistic_neg = torch.log(1 + torch.exp(neg_score)).sum(dim=1)
        logistic_pos = torch.log(1 + torch.exp(-pos_score)).sum(dim=1)
        logistic_loss = logistic_neg + logistic_pos

        re_p, im_p = self.model.rel_emb(self.rule_p).chunk(2, dim=-1)
        re_q, im_q = self.model.rel_emb(self.rule_q).chunk(2, dim=-1)
        entail_loss_re = self.args.mu * torch.sum(
            self.confidence * (re_p - re_q).clamp(min=0).sum(dim=-1)
        )
        entail_loss_im = self.args.mu * torch.sum(
            self.confidence * (im_p - im_q).pow(2).sum(dim=-1)
        )
        entail_loss = entail_loss_re + entail_loss_im
        loss = logistic_loss + entail_loss
        # return loss
        if self.args.regularization != 0.0:
            # Use L2 regularization for ComplEx_NNE_AER
            regularization = self.args.regularization * (
                self.model.ent_emb.weight.norm(p=2) ** 2
                + self.model.rel_emb.weight.norm(p=2) ** 2
            )
            loss = loss + regularization
        loss = loss.mean()
        return loss
